import struct
from typing import Any, BinaryIO, Union, Optional

from ofrak_type.endianness import Endianness


class SerializationError(Exception):
    pass


class BinarySerializer:
    def __init__(
        self,
        writer: Optional[BinaryIO] = None,
        endianness: Endianness = Endianness.BIG_ENDIAN,
        word_size: int = 8,
    ):
        self._writer = writer
        self._endianness = endianness
        self._word_size = word_size
        self._initial_position = 0 if writer is None else writer.tell()

    def set_stream(
        self,
        writer: BinaryIO,
        endianness: Endianness = Endianness.BIG_ENDIAN,
        word_size: int = 8,
    ):
        self._writer = writer
        self._endianness = endianness
        self._word_size = word_size
        self._initial_position = writer.tell()

    def set_endianness(self, endianness: Endianness) -> None:
        self._endianness = endianness

    def set_word_size(self, word_size: int) -> None:
        self._word_size = word_size

    def get_word_size(self) -> int:
        return self._word_size

    def seek(self, position: int) -> int:
        if self._writer is None:
            raise SerializationError("writer is not set")
        return self._writer.seek(self._initial_position + position)

    def position(self) -> int:
        if self._writer is None:
            raise SerializationError("writer is not set")
        return self._writer.tell()

    def write(self, data: bytes) -> int:
        if self._writer is None:
            raise SerializationError("writer is not set")
        length = self._writer.write(data)
        if len(data) != length:
            raise SerializationError(f"Could not write {len(data)} bytes")
        return length

    def pack_multiple(self, char: str, *values: Any, auto_bitwidth: bool = False) -> None:
        char = self._endianness.get_struct_flag() + char
        if auto_bitwidth and self._word_size != 8:
            char = char.replace("Q", "I").replace("q", "i")
        self.write(struct.pack(char, *values))

    def _pack(self, char: str, value: Union[int, float]) -> None:
        char = self._endianness.get_struct_flag() + char
        self.write(struct.pack(char, value))

    def pack_ubyte(self, value: int) -> None:
        self._pack("B", value)

    def pack_ushort(self, value: int) -> None:
        self._pack("H", value)

    def pack_uint(self, value: int) -> None:
        self._pack("I", value)

    def pack_ulong(self, value: int) -> None:
        char = "I" if self._word_size == 4 else "Q"
        self._pack(char, value)

    def pack_byte(self, value: int) -> None:
        self._pack("b", value)

    def pack_short(self, value: int) -> None:
        self._pack("h", value)

    def pack_int(self, value: int) -> None:
        self._pack("i", value)

    def pack_long(self, value: int) -> None:
        char = "i" if self._word_size == 4 else "q"
        self._pack(char, value)

    def pack_long_long(self, value: int) -> None:
        self._pack("q", value)

    def pack_float(self, value: float) -> None:
        self._pack("f", value)

    def pack_double(self, value: float) -> None:
        self._pack("d", value)

    def pack_string(self, value: str, length: int) -> None:
        value_raw = value.encode("utf-8")
        if len(value_raw) > length:
            raise SerializationError(f"The provided string does not fit in {length} bytes")
        self.write(value_raw.ljust(length, b"\x00"))

    def pack_dynamic_bytes(self, value: bytes) -> None:
        self.pack_ubyte(len(value))
        self.write(value)

    def pack_dynamic_bytes_short(self, value: bytes) -> None:
        self.pack_ushort(len(value))
        self.write(value)



"""
StreamCapture is useful utility for capturing standard output of C processes called from python.

**Usage:**
```python
with StreamCapture(sys.stdout) as stream_capture:
    sys.stdout.write("hello\\n")
assert stream_capture.get_captured_stream() == "hello\\n"
```
"""

from types import TracebackType
from typing import IO, Optional, Type

import os


class StreamCapture:
    """
    Capture a stream from a filelike object.
    """

    escape_char = b"\b"

    def __init__(self, stream: IO):
        self.stream = stream
        self.stream_file_descriptor = self.stream.fileno()
        self.pipe_out, self.pipe_in = os.pipe()
        self.stream_captured = ""
        self.stream_file_descriptor_copy = None  # type: Optional[int]

    def __enter__(self):
        self.start()
        return self

    def __exit__(
        self,
        exception_type: Optional[Type[BaseException]],
        exception_value: Optional[BaseException],
        exception_traceback: Optional[TracebackType],
    ) -> None:
        self.stop()

    def start(self) -> None:
        """
        Start capturing the stream.
        """
        self.stream_file_descriptor_copy = os.dup(self.stream_file_descriptor)
        os.dup2(self.pipe_in, self.stream_file_descriptor)

    def stop(self) -> None:
        """
        Stop capturing the stream and read what was captured.
        """
        if self.stream_file_descriptor_copy is None:
            # Capture was not started
            return

        # Flush and read the captured stream
        self.stream.write(self.escape_char.decode())
        self.stream.flush()
        self._read_stream()

        # Reset the file
        os.close(self.pipe_in)
        os.close(self.pipe_out)
        os.dup2(self.stream_file_descriptor_copy, self.stream_file_descriptor)
        os.close(self.stream_file_descriptor_copy)

        # Write the captured stream to the file. This way, anything that may have been captured
        # by accident still gets where it needs to go
        self.stream.write(self.stream_captured)

    def get_captured_stream(self) -> str:
        """
        Get the captured stream.
        """
        return self.stream_captured

    def _read_stream(self) -> None:
        """
        Read the stream data (one byte at a time)
        and save the text in `capturedtext`.
        """
        while True:
            char = os.read(self.pipe_out, 1)
            if not char or self.escape_char in char:
                break
            self.stream_captured += char.decode()



import io

import pytest
from ofrak_io.serializer import BinarySerializer, SerializationError

from ofrak_type.endianness import Endianness


class TestBinarySerializer:
    def test_set_stream(self):
        buffer = io.BytesIO()
        serializer = BinarySerializer()
        serializer.set_stream(buffer)
        serializer.write(b"hello")
        assert buffer.getvalue() == b"hello"

    def test_set_endianness(self):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)
        serializer.set_endianness(Endianness.LITTLE_ENDIAN)

    def test_set_word_size(self):
        serializer = BinarySerializer(io.BytesIO())
        assert serializer.get_word_size() == 8
        serializer.set_word_size(4)
        assert serializer.get_word_size() == 4

    def test_seek(self):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)
        assert serializer.position() == 0
        serializer.write(b"Hello")
        assert serializer.position() == len(b"Hello")
        assert buffer.getvalue() == b"Hello"
        serializer.seek(0)
        assert serializer.position() == 0
        serializer.write(b"world")
        assert serializer.position() == len(b"world")
        assert buffer.getvalue() == b"world"

    def test_seek_writer_not_set_error(self):
        serializer = BinarySerializer()
        with pytest.raises(SerializationError):
            serializer.seek(5)

    def test_position_writer_not_set_error(self):
        serializer = BinarySerializer()
        with pytest.raises(SerializationError):
            serializer.position()

    def test_write_error_writer_not_set(self):
        serializer = BinarySerializer()
        with pytest.raises(SerializationError):
            serializer.write(b"hello")

    def test_write_error_length(self):
        class OffByOneBuffer(io.BytesIO):
            pass

            def write(self, *args, **kwargs):  # real signature unknown
                first_arg = args[0]
                return len(first_arg) + 1

        buffer = OffByOneBuffer()
        serializer = BinarySerializer(buffer)
        with pytest.raises(SerializationError):
            serializer.write(b"hello")

    def test_pack_string_error(self):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)
        string_value = "hello"
        with pytest.raises(SerializationError):
            serializer.pack_string(string_value, len(string_value) - 1)



import os
import sys

from ofrak_io.stream_capture import StreamCapture


def test_stream_capture(capsys):
    """
    Test StreamCapture.
    """
    # pytest captures standard in and out by default. Disable this for our test
    with capsys.disabled():
        with StreamCapture(sys.stdout) as stream_capture:
            sys.stdout.write("hello\n")
        captured_stream = stream_capture.get_captured_stream()
        # On Windows, "\n" gets converted to "\r\n" (linesep)
        assert captured_stream == f"hello{os.linesep}"


def test_stream_capture_stop_called_before_start(capsys):
    with capsys.disabled():
        stream = StreamCapture(sys.stdout)
        result = stream.stop()
        assert result is None



import io
import os
import random
import string
from dataclasses import dataclass

from hypothesis import given, strategies
from ofrak_io.deserializer import BinaryDeserializer
from ofrak_io.serializer import BinarySerializer

ubyte_strategy = strategies.integers(min_value=0, max_value=2**8 - 1)
uint16_strategy = strategies.integers(min_value=0, max_value=2**16 - 1)
uint32_strategy = strategies.integers(min_value=0, max_value=2**32 - 1)
uint64_strategy = strategies.integers(min_value=0, max_value=2**64 - 1)

signed_byte_strategy = strategies.integers(min_value=-(2**7), max_value=2**7 - 1)
signed_int16_strategy = strategies.integers(min_value=-(2**15), max_value=2**15 - 1)
signed_int32_strategy = strategies.integers(min_value=-(2**31), max_value=2**31 - 1)
signed_int64_strategy = strategies.integers(min_value=-(2**63), max_value=2**63 - 1)
long_long_strategy = strategies.integers(min_value=-(2**63), max_value=2**63 - 1)
float_strategy = strategies.floats(width=32, allow_nan=False)
double_float_strategy = strategies.floats(width=64, allow_nan=False)
string_strategy = strategies.text(alphabet=string.printable)


@dataclass
class PackMultipleValues:
    i_value: int
    b_value: int
    q_value: int

    @property
    def char(self):
        return "IBQ"


pack_multiple_strategy = strategies.builds(
    PackMultipleValues, uint16_strategy, ubyte_strategy, uint32_strategy
)


class TestBinarySerializationIdentity:
    """
    Test that serializing a value with the `BinarySerializer`, and then deserializing with the
    `BinaryDeserializer` produces the same value ("identity" test).
    """

    @given(values=pack_multiple_strategy)
    def test_multiple(self, values: PackMultipleValues):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)
        serializer.pack_multiple(values.char, values.i_value, values.b_value, values.q_value)
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized))
        deserialized = deserializer.unpack_multiple(values.char)
        assert PackMultipleValues(*deserialized) == values

    @given(values=pack_multiple_strategy)
    def test_multiple_auto_bitwidth(self, values: PackMultipleValues):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer, word_size=4)
        serializer.pack_multiple(
            values.char, values.i_value, values.b_value, values.q_value, auto_bitwidth=True
        )
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized), word_size=4)
        deserialized = deserializer.unpack_multiple(values.char, auto_bitwidth=True)
        assert PackMultipleValues(*deserialized) == values

    @given(value=ubyte_strategy)
    def test_ubyte(self, value: int):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)

        serializer.pack_ubyte(value)
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized))
        deserialized = deserializer.unpack_ubyte()
        assert deserialized == value

    @given(value=uint16_strategy)
    def test_ushort(self, value: int):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)

        serializer.pack_ushort(value)
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized))
        deserialized = deserializer.unpack_ushort()
        assert deserialized == value

    @given(value=uint32_strategy)
    def test_uint(self, value: int):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)

        serializer.pack_uint(value)
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized))
        deserialized = deserializer.unpack_uint()
        assert deserialized == value

    @given(value=uint64_strategy)
    def test_ulong(self, value: int):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)

        serializer.pack_ulong(value)
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized))
        deserialized = deserializer.unpack_ulong()
        assert deserialized == value

    @given(value=signed_byte_strategy)
    def test_byte(self, value: int):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)

        serializer.pack_byte(value)
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized))
        deserialized = deserializer.unpack_byte()
        assert deserialized == value

    @given(value=signed_int16_strategy)
    def test_short(self, value: int):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)

        serializer.pack_short(value)
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized))
        deserialized = deserializer.unpack_short()
        assert deserialized == value

    @given(value=signed_int32_strategy)
    def test_int(self, value: int):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)

        serializer.pack_int(value)
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized))
        deserialized = deserializer.unpack_int()
        assert deserialized == value

    @given(value=signed_int64_strategy)
    def test_long(self, value: int):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)

        serializer.pack_long(value)
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized))
        deserialized = deserializer.unpack_long()
        assert deserialized == value

    @given(value=long_long_strategy)
    def test_long_long(self, value: int):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)

        serializer.pack_long_long(value)
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized))
        deserialized = deserializer.unpack_long_long()
        assert deserialized == value

    @given(value=float_strategy)
    def test_float(self, value: float):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)
        serializer.pack_float(value)
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized))
        deserialized = deserializer.unpack_float()
        assert deserialized == value

    @given(value=double_float_strategy)
    def test_double(self, value: float):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)
        serializer.pack_double(value)
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized))
        deserialized = deserializer.unpack_double()
        assert deserialized == value

    @given(value=string_strategy)
    def test_string(self, value: str):
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)
        string_length = len(value) + random.randint(0, 4)
        serializer.pack_string(value, string_length)
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized))
        deserialized = deserializer.unpack_string(len(value))
        assert deserialized == value

    def test_variable_length_string(self):
        string_text = "hello"
        string_length = len(string_text) + 10
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)
        serializer.pack_string(string_text, string_length)
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized))
        deserialized = deserializer.unpack_string()
        assert deserialized == string_text

    @given(value=ubyte_strategy)
    def test_dynamic_bytes(self, value: int):
        raw_bytes = os.urandom(value)
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)

        serializer.pack_dynamic_bytes(raw_bytes)
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized))
        deserialized = deserializer.unpack_dynamic_bytes()
        assert deserialized == raw_bytes

    @given(value=uint16_strategy)
    def test_dynamic_bytes_short(self, value: int):
        raw_bytes = os.urandom(value)
        buffer = io.BytesIO()
        serializer = BinarySerializer(buffer)

        serializer.pack_dynamic_bytes_short(raw_bytes)
        serialized = buffer.getvalue()

        deserializer = BinaryDeserializer(io.BytesIO(serialized))
        deserialized = deserializer.unpack_dynamic_bytes_short()
        assert deserialized == raw_bytes



import io

import pytest
from ofrak_io.deserializer import BinaryDeserializer, DeserializationError

from ofrak_type.endianness import Endianness


class TestBinaryDeserializer:
    def test_set_stream(self):
        message = b"hello"
        deserializer = BinaryDeserializer()
        deserializer.set_stream(io.BytesIO(message))
        deserialized = deserializer.read(len(message))
        assert deserialized == message

    def test_set_endianness(self):
        deserializer = BinaryDeserializer(io.BytesIO())
        deserializer.set_endianness(Endianness.LITTLE_ENDIAN)

    def test_set_word_size(self):
        deserializer = BinaryDeserializer(io.BytesIO())
        assert deserializer.get_word_size() == 8
        deserializer.set_word_size(4)
        assert deserializer.get_word_size() == 4

    def test_seek(self):
        message = b"HelloWorld"
        deserializer = BinaryDeserializer(io.BytesIO(message))
        assert deserializer.position() == 0
        deserializer.seek(len(b"Hello"))
        assert deserializer.position() == len(b"Hello")
        assert deserializer.read(len(b"World")) == b"World"
        deserializer.seek(0)
        assert deserializer.position() == 0
        assert deserializer.read(len(b"Hello")) == b"Hello"
        assert deserializer.position() == len(b"World")

    def test_seek_error(self):
        deserializer = BinaryDeserializer()
        with pytest.raises(DeserializationError):
            deserializer.seek(5)

    def test_position_error(self):
        deserializer = BinaryDeserializer()
        with pytest.raises(DeserializationError):
            deserializer.position()

    def test_read_error(self):
        deserializer = BinaryDeserializer()
        with pytest.raises(DeserializationError):
            deserializer.read(5)

    def test_read_length_error(self):
        deserializer = BinaryDeserializer(io.BytesIO(b""))
        with pytest.raises(DeserializationError):
            deserializer.read(10)



import struct
from typing import cast, Union, Tuple, BinaryIO, Optional

from ofrak_type.endianness import Endianness


class DeserializationError(Exception):
    pass


class BinaryDeserializer:
    def __init__(
        self,
        reader: Optional[BinaryIO] = None,
        endianness: Endianness = Endianness.BIG_ENDIAN,
        word_size: int = 8,
    ):
        self._reader = reader
        self._endianness = endianness
        self._word_size = word_size
        self._initial_position = 0 if reader is None else reader.tell()

    def set_stream(
        self,
        reader: BinaryIO,
        endianness: Endianness = Endianness.BIG_ENDIAN,
        word_size: int = 8,
    ):
        self._reader = reader
        self._endianness = endianness
        self._word_size = word_size
        self._initial_position = reader.tell()

    def set_endianness(self, endianness: Endianness) -> None:
        self._endianness = endianness

    def set_word_size(self, word_size: int) -> None:
        self._word_size = word_size

    def get_word_size(self) -> int:
        return self._word_size

    def seek(self, position: int) -> int:
        if self._reader is None:
            raise DeserializationError("reader is not set")
        return self._reader.seek(self._initial_position + position)

    def position(self) -> int:
        if self._reader is None:
            raise DeserializationError("reader is not set")
        return self._reader.tell()

    def read(self, length: int) -> bytes:
        if self._reader is None:
            raise DeserializationError("reader is not set")
        data = self._reader.read(length)
        if len(data) != length:
            raise DeserializationError(f"Could not read {length} bytes, data len is {len(data)}")
        return data

    def unpack_multiple(self, char: str, length: int = -1, auto_bitwidth: bool = False) -> Tuple:
        char = self._endianness.get_struct_flag() + char
        if auto_bitwidth and self._word_size != 8:
            char = char.replace("Q", "I").replace("q", "i")
        if length <= 0:
            length = struct.calcsize(char)
        return struct.unpack(char, self.read(length))

    def _unpack(self, char: str, length: int) -> Union[int, float]:
        char = self._endianness.get_struct_flag() + char
        (result,) = struct.unpack(char, self.read(length))
        return result

    def unpack_ubyte(self) -> int:
        return cast(int, self._unpack("B", 1))

    def unpack_ushort(self) -> int:
        return cast(int, self._unpack("H", 2))

    def unpack_uint(self) -> int:
        return cast(int, self._unpack("I", 4))

    def unpack_ulong(self) -> int:
        char = "I" if self._word_size == 4 else "Q"
        return cast(int, self._unpack(char, self._word_size))

    def unpack_byte(self) -> int:
        return cast(int, self._unpack("b", 1))

    def unpack_short(self) -> int:
        return cast(int, self._unpack("h", 2))

    def unpack_int(self) -> int:
        return cast(int, self._unpack("i", 4))

    def unpack_long(self) -> int:
        char = "i" if self._word_size == 4 else "q"
        return cast(int, self._unpack(char, self._word_size))

    def unpack_long_long(self) -> int:
        return cast(int, self._unpack("q", 8))

    def unpack_float(self) -> float:
        return cast(float, self._unpack("f", 4))

    def unpack_double(self) -> float:
        return cast(float, self._unpack("d", 8))

    def unpack_string(self, length: int = -1) -> str:
        if length < 0:
            current_string = []
            while True:
                current_char = self.read(1)
                if current_char == b"\x00":
                    break
                else:
                    current_string.append(current_char.decode("utf-8"))
            value = "".join(current_string)
        else:
            value = self.read(length).replace(b"\x00", b"").decode("utf-8")
        return value

    def unpack_dynamic_bytes(self) -> bytes:
        length = self.unpack_ubyte()
        return self.read(length)

    def unpack_dynamic_bytes_short(self) -> bytes:
        length = self.unpack_ushort()
        return self.read(length)



import asyncio
from abc import ABC, abstractmethod
from typing import TypeVar, Generic, Callable, Dict, Awaitable, Iterable, Tuple, ClassVar

Request = TypeVar("Request")
Result = TypeVar("Result")


_RequestKeyT = str
_BatchHandlerFunctionT = Callable[
    [Tuple[Request, ...]], Awaitable[Iterable[Tuple[Request, Result]]]
]
_DEFAULT_RATE_LIMIT = 10


def _DEFAULT_REQUEST_KEY(req):
    return str(hash(req))


class NotAllRequestsHandledError(Exception):
    pass


class BatchManagerInterface(Generic[Request, Result], ABC):
    """
    Class which manages automatically batching async requests to some resource (like a remote
    server) to limit the number of individual requests.
    """

    @abstractmethod
    async def get_result(self, request: Request) -> Result:
        """
        Get the result for a request. The request may be batched with one or more other pending
        requests before being passed to the `handler_function`.

        :param request: request to be passed in the argument of `handler_function`

        :return: result for the given request

        :raises NotAllRequestsHandledError: if this or any other requests were not handled by the
        `handler_function` passed to the constructor.
        """
        raise NotImplementedError()


def make_batch_manager(
    handler_function: _BatchHandlerFunctionT,
    rate_limit: int = _DEFAULT_RATE_LIMIT,
    request_key_f: Callable[[Request], _RequestKeyT] = _DEFAULT_REQUEST_KEY,
) -> BatchManagerInterface[Request, Result]:
    """
    Construct an object which will automatically batch every call to `get_result` into periodic
    calls to `handler_function`.

    This function is the preferred way to make a one-off batch manager with minimal lines of
    code. If you find yourself calling this function with the same arguments multiple times,
    consider instead defining a subclass of `AbstractBatchManager`. This is functionally
    equivalent to calling `make_batch_manager` with the same arguments, but the code is cleaner.

    The returned BatchManagerInterface is implemented for asyncio, but is NOT GENERALLY
    THREAD-SAFE! That is, it must not be shared between manually accessed threads.

    :param handler_function: function to handle multiple requests at once and return the results
    as pairs of (request, result)
    :param rate_limit:  maximum number of times `handler_function` will be called per second
    :param request_key_f: function this manager should use to uniquely identify requests

    :return: an instance of a `BatchManagerInterface` using `handler_function` to get results
    """
    return _BatchManagerImplementation(handler_function, rate_limit, request_key_f)


class _BatchManagerImplementation(BatchManagerInterface[Request, Result]):
    def __init__(
        self,
        handler_function: _BatchHandlerFunctionT,
        rate_limit: int,
        request_key_f: Callable[[Request], _RequestKeyT],
    ):
        # Basic state setup
        self._request_key_f = request_key_f
        self._handler_function = handler_function
        self._rate_limit = rate_limit
        # Background task and batch setup
        loop = asyncio.get_event_loop()
        self._handler_loop_task = loop.create_task(self._periodic_batch_handler())
        self._current_batch = self._new_batch()

    async def get_result(self, request: Request) -> Result:
        current_batch = self._current_batch
        current_batch.add_request(request)
        # Gives self._handler_loop_task a chance to raise its errors
        done, _ = await asyncio.wait(
            (asyncio.create_task(current_batch.result(request)), self._handler_loop_task),
            return_when=asyncio.FIRST_COMPLETED,
        )
        return next(iter(done)).result()

    async def _periodic_batch_handler(self):
        while True:
            await asyncio.sleep(1.0 / float(self._rate_limit))
            old_batch = self._current_batch
            if old_batch.has_requests():
                self._current_batch = self._new_batch()
                handled_results = await self._handler_function(old_batch.get_requests())
                old_batch.resolve_batch_requests(handled_results)

    def _new_batch(self):
        return _Batch[Request, Result](self._request_key_f)


class AbstractBatchManager(_BatchManagerImplementation[Request, Result], ABC):
    """
    An implementation of `BatchManagerInterface` which allows for defining a pattern of
    batch managers as a subclass. See that `BatchManagerInterface`'s documentation for details on
    what this class does. This class is a Generic type with two type arguments, for the type of
    each individual request and each individual return type.

    Subclassing AbstractBatchManager is preferred when `make_batch_manager` is called in multiple
    places with the same arguments, or if the handler function needs some persistent state. For
    making a one-off batch manager with minimal lines of code, use `make_batch_manager`.

    This class is implemented for asyncio, but is NOT GENERALLY THREAD-SAFE! That is, instances
    must not be shared between manually accessed threads.

    :cvar rate_limit: maximum number of times `handle_requests` will be called per second
    (equivalent to `rate_limit` argument to `make_batch_manager`)
    """

    rate_limit: ClassVar[int] = _DEFAULT_RATE_LIMIT

    def __init__(self):
        super().__init__(self.handle_requests, self.rate_limit, self.unique_key_for_request)

    @abstractmethod
    async def handle_requests(
        self, requests: Tuple[Request, ...]
    ) -> Iterable[Tuple[Request, Result]]:
        """
        Handle multiple requests at once and return the results as pairs of (request, result).
        Equivalent to `handler_function` argument to `make_batch_manager`.

        :param requests: one or more objects that were passed to `get_result`

        :return: pairs of (request, result)
        """
        raise NotImplementedError()

    @staticmethod
    def unique_key_for_request(request: Request) -> _RequestKeyT:
        """
        Function to uniquely identify requests. If the Request type is unhashable, this method
        must be overridden (the default function uses the hash).

        Equivalent to `request_key_f` argument to `make_batch_manager`.
        """
        return _DEFAULT_REQUEST_KEY(request)


class _Batch(Generic[Request, Result]):
    def __init__(self, request_key):
        self._request_key = request_key
        self._batch_was_handled = asyncio.Event()
        self._unresolved_requests: Dict[_RequestKeyT, Request] = dict()
        self._resolved_requests: Dict[_RequestKeyT, Result] = dict()

    def add_request(self, request: Request):
        self._unresolved_requests[self._request_key(request)] = request

    def has_requests(self) -> bool:
        return not len(self._unresolved_requests) == 0

    def get_requests(self) -> Tuple[Request, ...]:
        return tuple(self._unresolved_requests.values())

    async def result(self, request: Request) -> Result:
        await self._batch_was_handled.wait()
        return self._resolved_requests[self._request_key(request)]

    def resolve_batch_requests(self, request_result_pairs: Iterable[Tuple[Request, Result]]):
        for req, res in request_result_pairs:
            k = self._request_key(req)
            self._resolved_requests[k] = res
            self._unresolved_requests.pop(k)

        if len(self._unresolved_requests) != 0:
            raise NotAllRequestsHandledError(
                f"Could not get the results for this batch because "
                f"{len(self._unresolved_requests)} requests were unhandled!"
            )
        self._batch_was_handled.set()



import asyncio
import random
from dataclasses import dataclass
from typing import Tuple, Iterable

import pytest

from ofrak_io.batch_manager import (
    NotAllRequestsHandledError,
    AbstractBatchManager,
    make_batch_manager,
)


@dataclass
class Counter:
    count: int


@pytest.fixture
def many_strings():
    def random_string():
        return "".join([chr(random.choice(range(0, 256))) for _ in range(random.randint(3, 15))])

    return [random_string() for _ in range(1000)]


@pytest.fixture
def counter():
    return Counter(0)


def batch_manager_with_function(counter):
    async def handle_requests(requests: Tuple[str, ...]) -> Iterable[Tuple[str, int]]:
        await asyncio.sleep(1)
        results = [(req, len(req)) for req in requests]
        counter.count += 1
        return results

    return make_batch_manager(handle_requests)


def batch_manager_with_subclass(counter):
    class TestBatchManager(AbstractBatchManager[str, int]):
        async def handle_requests(self, requests: Tuple[str, ...]) -> Iterable[Tuple[str, int]]:
            await asyncio.sleep(1)
            results = [(req, len(req)) for req in requests]
            counter.count += 1
            return results

    return TestBatchManager()


@pytest.fixture(params=(batch_manager_with_function, batch_manager_with_subclass))
def batch_manager(request, counter):
    return request.param(counter)


@pytest.fixture
def bad_batch_manager(counter):
    async def handle_requests(requests: Tuple[str, ...]) -> Iterable[Tuple[str, int]]:
        await asyncio.sleep(1)
        return []

    return make_batch_manager(handle_requests)


async def test_single_result(batch_manager, counter):
    test_string = "OFRAk is awesome"
    res = await batch_manager.get_result(test_string)
    assert res == len(test_string)
    assert counter.count == 1


async def test_many_results(batch_manager, many_strings, counter):
    await asyncio.gather(*[batch_manager.get_result(s) for s in many_strings])
    assert counter.count == 1


async def test_incomplete_handling_raises_err(bad_batch_manager):
    test_string = "OFRAk is awesome"
    with pytest.raises(NotAllRequestsHandledError):
        _ = await bad_batch_manager.get_result(test_string)
